from subprocess import Popen, PIPE, DEVNULL
import sys

# There are 3+ command line arguments:
# 1) the file to update
# 2) the file's previous state, to check if things changed at all
# 3) the file's parents that need to be updated as well

FILE = sys.argv[1]
OLD_STATE = sys.argv[2]
PARENTS = [p for p in sys.argv[3:]]

FILE_STATE_CMD = "git status --porcelain --ignored "
IS_IGNORED_CMD = "git check-ignore "
IS_TRACKED_CMD = "git ls-files --error-unmatch "
IS_CHANGED_CMD = "git diff-index --quiet HEAD "

def main():
    new_state = determine_file_git_state()

    # nothing to do
    if OLD_STATE == new_state:
        sys.exit(2)

    proc_list = []

    # for every parent file start all necessary git processes immediately
    # even if we don't need them later
    for p in PARENTS:
        add_git_processes(proc_list, p)

    result_list = [(FILE, new_state)]

    # iterate through the parents and propagate ignored and untracked states downwards
    # the following states are possible for *directories*:
    # 0 -> clean
    # ! -> ignored
    # ? -> untracked
    # M -> modified
    i = 0
    l = len(proc_list)
    propagate_state = None
    while i < l:
        path, ignore_proc, tracked_proc, changed_proc = proc_list[i]
        if ignore_proc.wait() == 0:
            propagate_state = "!"
            result_list.append((path, propagate_state))
            break
        elif tracked_proc.wait() == 1:
            propagate_state = "?"
            result_list.append((path, propagate_state))
            break
        elif changed_proc.wait() == 1:
            result_list.append((path, "M"))
        else:
            result_list.append((path, "0"))
        i += 1

    if propagate_state:
        i += 1
        while i < l:
            result_list.append((proc_list[i][0], propagate_state))
            i += 1

    elisp_conses = "".join(['("{}" . "{}")'.format(path, state) for path,state in result_list])
    elisp_alist = "({})".format(elisp_conses)
    print(elisp_alist)

def add_git_processes(status_listings, path):
    ignored_proc = Popen(IS_IGNORED_CMD + path, shell=True, stdout=DEVNULL, stderr=DEVNULL)
    tracked_proc = Popen(IS_TRACKED_CMD + path, shell=True, stdout=DEVNULL, stderr=DEVNULL)
    changed_proc = Popen(IS_CHANGED_CMD + path, shell=True, stdout=DEVNULL, stderr=DEVNULL)

    status_listings.append((path, ignored_proc, tracked_proc, changed_proc))

def determine_file_git_state():
    proc  = Popen(FILE_STATE_CMD + FILE, shell=True, stdout=PIPE, stderr=DEVNULL)
    line  = proc.stdout.readline()
    if line:
        state = line.lstrip().split(b" ")[0]
        return state.decode('utf-8').strip()[0]
    else:
        return "0"

main()



from subprocess import Popen, PIPE
from os import listdir, environ
from os.path import isdir, islink
from posixpath import join
import sys

# The script is supplied with the followig command line arguments
# 1) the repository root - used only for file-joining to an absolute path
#    the actual working directory is set in emacs
# 2) `treemacs-max-git-entries`
# 3) `treemacs-git-command-pipe`
# 4) a list of expanded directories the script may recurse into to collect
#    an entry for every untracked/ignored file inside
#    this list is turned into a set since it is possible that it contains duplicates
#    when called for magit, see also `treemacs-magit--extended-git-mode-update`

GIT_ROOT     = str.encode(sys.argv[1])
LIMIT        = int(sys.argv[2])
GIT_CMD      = "git status --porcelain --ignored . " + sys.argv[3]
STDOUT       = sys.stdout.buffer
RECURSE_DIRS = set([str.encode(it[(len(GIT_ROOT)):]) + b"/" for it in sys.argv[4:]]) if len(sys.argv) > 4 else []
QUOTE        = b'"'
output       = []
ht_size      = 0

def find_recursive_entries(path, state):
    global output, ht_size
    for item in listdir(path):
        full_path = join(path, item)
        output.append(full_path + QUOTE + QUOTE + state)
        ht_size += 1
        if ht_size > LIMIT:
            break
        if isdir(full_path) and not islink(full_path):
            find_recursive_entries(full_path, state)

def main():
    global output, ht_size
    # Don't lock Git when updating status.
    environ["GIT_OPTIONAL_LOCKS"] = "0"
    proc = Popen(GIT_CMD, shell=True, stdout=PIPE, bufsize=100)
    dirs_added = {}

    for item in proc.stdout:
        # remove final newline
        item = item[:-1]

        # remove leading space if item was e.g. modified only in the worktree
        if item.startswith(b' '):
            item = item[1:]
        state, filename = item.split(b' ', 1)

        # reduce the state to a single-letter-string
        state = state[0:1]

        # sometimes git outputs quoted filesnames
        if filename.startswith(b'"'):
            filename = filename[1:-1]

        # find the absolute path for the current item
        # renames have the form STATE OLDNAME -> NEWNAME
        abs_path = None
        if state == b"R":
            abs_path = join(GIT_ROOT, filename.split(b' -> ')[1])
        else:
            abs_path = join(GIT_ROOT, filename.lstrip())

        # filename is a directory, final slash must be removed
        if abs_path.endswith(b'/'):
            abs_path = abs_path[:-1]
            dirs_added[abs_path] = True
        output.append(abs_path + QUOTE + QUOTE + state)
        ht_size += 1

        # for files deeper down in the file hierarchy also print all their directories
        # if /A/B/C/x is changed then /A and /A/B and /A/B/C must be shown as changed as well
        if b'/' in filename and state != b'!':
            name_parts = filename.split(b'/')[:-1]
            dirname = b''
            for name_part in name_parts:
                dirname = join(dirname, name_part)
                full_dirname = join(GIT_ROOT, dirname.lstrip())
                # directories should not be printed more than once, which would happen if
                # e.g. both /A/B/C/x and /A/B/C/y have changes
                if full_dirname not in dirs_added:
                    output.append(full_dirname + QUOTE + QUOTE + b'M')
                    ht_size += 1
                    dirs_added[full_dirname] = True
        # for untracked and ignored directories we need to find an entry for every single file
        # they contain
        # however this applies only for directories that are expanded and whose content is visible
        if state in [b'?', b'!'] and isdir(abs_path):
            if filename in RECURSE_DIRS:
                find_recursive_entries(abs_path, state)
        if ht_size >= LIMIT:
            break
    STDOUT.write(
        b"#s(hash-table size " + \
        bytes(str(ht_size), 'utf-8') + \
        b" test equal rehash-size 1.5 rehash-threshold 0.8125 data ("
    )
    if ht_size > 0:
      STDOUT.write(QUOTE + (QUOTE + QUOTE).join(output) + QUOTE)
    STDOUT.write( b"))")

    sys.exit(proc.poll())

main()



from os import listdir
from os.path import isdir
from posixpath import join
import sys
import os

ROOT     = sys.argv[1]
LIMIT    = int(sys.argv[2])
SHOW_ALL = sys.argv[3] == 't'

# special workaround for windows platforms
# the default `join' implementation cannot quite deal with windows
# paths in the form of "C:/A/B" & "C:/A/B/C", joining them as
# "C:/A/B/C:/A/B/C"
# it can, however, be "tricked" into doing the right thing by adding
# a slash to the start of the paths
# go figure
if sys.platform == 'win32':
    def join_dirs(d1, d2, full_path=False):
        missing_slash = False
        if not d1.startswith("/"):
            missing_slash = True
            d1 = "/" + d1
        # full_path is only True when the second argument is
        # another absolute path
        if full_path and not d2.startswith("/"):
            missing_slash = True
            d2 = "/" + d2
        joined = join(d1, d2)
        if missing_slash:
            # still need to return the joined path without the
            # leading slash, the way it looked originally
            return joined[1:]
        else:
            return joined
else:
    def join_dirs(d1, d2, *_):
        return join(d1, d2)

if LIMIT <= 0:
    exit(0)

def dir_content(path):
    """
    returns the content of given path, excluding unreadable files
    and dotfiles (unless SHOW_ALL is True)
    """
    ret = []
    for item in listdir(path):
        full_path = join_dirs(path, item)
        if os.access(full_path, os.R_OK) and (SHOW_ALL or item[0] != '.'):
            ret.append(full_path)
    return ret

def main():
    out  = sys.stdout
    dirs = [d for d in dir_content(ROOT) if isdir(d)]
    out.write("(")
    for current_dir in dirs:
        content   = dir_content(current_dir)
        collapsed = current_dir
        steps     = []
        depth     = 0
        while True:
            if len(content) == 1 and isdir(content[0]):
                single_path = content[0]
                collapsed   = join_dirs(collapsed, single_path, True)
                content     = dir_content(collapsed)
                depth      += 1
                steps.append(single_path)
                if depth >= LIMIT:
                    break
            else:
                break
        if depth > 0:
            final_dir      = steps[-1]
            display_suffix = final_dir[len(current_dir):]
            out.write("(" + '"' + display_suffix + '" ' + '"' + current_dir + '" ' + '"' + '" "'.join(steps) + '")')
    out.write(")")

main()
